"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt

from mypo.common import safe_cast
from mypo.trigger.base_trigger import BaseTrigger


class BaseRebalancer(object):
    """Interface class of Rebalance stragegy class."""

    _trigger: BaseTrigger

    def __init__(self, trigger: BaseTrigger, weights: npt.ArrayLike) -> None:
        self._trigger = trigger
        self._weights = safe_cast(weights)

    def apply(
        self, index: datetime.datetime, assets: npt.ArrayLike, cash: np.float64
    ) -> np.ndarray:
        """
        Apply rebalance strategy to current situation.

        Parameters
        ----------
        index
            Current date for applying rebalance.

        assets
            Current assets for applying rebalance.

        cash
            Current cash for applying rebalance.

        Returns
        -------
        Deal
        """
        if self._trigger.is_fire(index, assets, cash, self._weights):
            return self._rebalance(safe_cast(assets))
        else:
            return self._do_nothing(safe_cast(assets))

    def _rebalance(self, assets: np.ndarray) -> np.ndarray:
        diff: np.ndarray = self._weights * np.sum(assets) - assets
        return diff

    def _do_nothing(self, assets: np.ndarray) -> np.ndarray:
        zero: np.ndarray = assets - assets
        return zero



"""Rebalance strategies."""

import numpy as np
import numpy.typing as npt

from mypo.rebalancer.base_rebalancer import BaseRebalancer
from mypo.trigger.threshold_trigger import ThresholdTrigger


class ThresholdRebalancer(BaseRebalancer):
    """Weighted rebalance strategy by monthly applying."""

    _threshold: np.float64

    def __init__(
        self, weights: npt.ArrayLike, threshold: np.float64 = np.float64(0.05)
    ) -> None:
        """
        Construct object.

        Parameters
        ----------
        weights
            Weight for applying rebalance.

        threshold
            Threshold of fire.

        """
        super().__init__(trigger=ThresholdTrigger(threshold=threshold), weights=weights)



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt

from mypo.common import safe_cast
from mypo.trigger.base_trigger import BaseTrigger


class MonthlyTrigger(BaseTrigger):
    """Weighted rebalance strategy by monthly applying."""

    _old_month: int
    _weights: np.ndarray

    def __init__(self, old_month: int = 0) -> None:
        """
        Construct object.

        Parameters
        ----------
        old_month
            Previous month.
        """
        super().__init__()
        self._old_month = old_month

    def is_fire(
        self,
        index: datetime.datetime,
        assets: npt.ArrayLike,
        cash: np.float64,
        weights: npt.ArrayLike,
    ) -> bool:
        """
        Apply rebalance strategy to current situation.

        Parameters
        ----------
        index
            Current date for applying rebalance.

        assets
            Current assets for applying rebalance.

        cash
            Current cash for applying rebalance.

        weights
            Weights of assets.

        Returns
        -------
        Deal
        """
        assets = safe_cast(assets)
        if self._old_month != index.month:
            self._old_month = index.month
            return True
        else:
            return False



"""Rebalance strategies."""
import numpy.typing as npt

from mypo.rebalancer.base_rebalancer import BaseRebalancer
from mypo.trigger.always_trigger import AlwaysTrigger


class PlainRebalancer(BaseRebalancer):
    """Simple weighted rebalance strategy."""

    def __init__(self, weights: npt.ArrayLike) -> None:
        """
        Construct object.

        Parameters
        ----------
        weights
            Weight for applying rebalance.
        """
        super().__init__(trigger=AlwaysTrigger(), weights=weights)



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt


class BaseTrigger(object):
    """Weighted rebalance strategy by monthly applying."""

    def is_fire(
        self,
        index: datetime.datetime,
        assets: npt.ArrayLike,
        cash: np.float64,
        weights: npt.ArrayLike,
    ) -> bool:
        """
        Apply rebalance strategy to current situation.

        Parameters
        ----------
        index
            Current date for applying rebalance.

        assets
            Current assets for applying rebalance.

        cash
            Current cash for applying rebalance.

        weights
            Weight of assets.

        Returns
        -------
        Deal
        """
        pass



"""Optimizer for weights of portfolio."""

import numpy as np
import pandas as pd
from scipy.optimize import minimize

from mypo.common import safe_cast
from mypo.market import Market
from mypo.optimizer import Optimizer
from mypo.optimizer.objective import sharp_ratio


class SharpRatioOptimizer(Optimizer):
    """Minimum variance optimizer."""

    _historical_data: pd.DataFrame
    _span: int

    def __init__(
        self,
        market: Market,
        risk_free_rate: np.float64 = np.float64(0.02),
        span: int = 260,
    ):
        """
        Construct this object.

        Parameters
        ----------
        market
            Past market stock prices.
        risk_free_rate
            Risk free rate

        span
            Span for evaluation.
        """
        self._historical_data = market.get_rate_of_change()
        self._risk_free_rate = risk_free_rate
        self._span = span

    def optimize_weight(self) -> np.ndarray:
        """
        Optimize weights.

        Returns
        -------
        Optimized weights
        """
        prices = self._historical_data.tail(n=self._span).to_numpy()
        Q = np.cov(prices.T)
        R = prices.mean(axis=0).T
        n = Q.shape[0]
        x = np.ones(n) / n
        daily_risk_free_rate = (1.0 + self._risk_free_rate) ** (1 / 252) - 1.0

        def fn(
            x: np.ndarray,
            R: np.ndarray,
            Q: np.ndarray,
            daily_risk_free_rate: np.float64,
        ) -> np.float64:
            return -sharp_ratio(
                np.dot(x, R), np.dot(np.dot(x, Q), x.T), daily_risk_free_rate
            )

        cons = [{"type": "eq", "fun": lambda x: np.sum(x) - 1}]
        bounds = [[0.0, 1.0] for i in range(n)]
        minout = minimize(
            fn,
            x,
            args=(R, Q, daily_risk_free_rate),
            method="SLSQP",
            bounds=bounds,
            constraints=cons,
        )
        return safe_cast(minout.x)



"""Rebalance strategies."""

import numpy as np
import numpy.typing as npt

from mypo.common import safe_cast
from mypo.rebalancer.base_rebalancer import BaseRebalancer
from mypo.trigger.monthly_trigger import MonthlyTrigger


class MonthlyRebalancer(BaseRebalancer):
    """Weighted rebalance strategy by monthly applying."""

    _old_month: int
    _weights: np.ndarray

    def __init__(self, weights: npt.ArrayLike, old_month: int = 0) -> None:
        """
        Construct object.

        Parameters
        ----------
        weights
            Weight for applying rebalance.

        old_month
            Previous month.
        """
        super().__init__(trigger=MonthlyTrigger(old_month), weights=weights)
        self._old_month = old_month
        self._weights = safe_cast(weights)



"""Covariance and semi-covariance."""
from typing import Callable

import numpy as np

CovarianceModel = Callable[[np.ndarray], np.ndarray]


def covariance(prices: np.ndarray) -> np.ndarray:
    """
    Calculate covariance matrix.

    Parameters
    ----------
    prices
        Prices of market data.

    Returns
    -------
    Covariance matrix.

    """
    Q = np.cov(prices.T)
    return np.array(Q)


def semi_covariance(prices: np.ndarray) -> np.ndarray:
    """
    Calculate semi-covariance matrix.

    Parameters
    ----------
    prices
        Prices of market data.

    Returns
    -------
    Semi-covariance matrix.

    """
    Q = np.cov(np.where(prices.T < 0, prices.T, 0.0))
    return np.array(Q)


def sharp_ratio(r: np.float64, q: np.float64, risk_free_rate: np.float64) -> np.float64:
    """
    Calculate Sharp ratio.

    Parameters
    ----------
    r
        Return
    q
        Variance
    risk_free_rate
        Risk free rate

    Returns
    -------
    Sharp ratio.

    """
    return (r - risk_free_rate) / q



"""Rebalance strategies."""
import datetime

import numpy as np
import numpy.typing as npt

from mypo.trigger.base_trigger import BaseTrigger


class NoTrigger(BaseTrigger):
    """Never fire trigger."""

    def __init__(self) -> None:
        """Construct object."""
        super().__init__()

    def is_fire(
        self,
        index: datetime.datetime,
        assets: npt.ArrayLike,
        cash: np.float64,
        weights: npt.ArrayLike,
    ) -> bool:
        """
        Apply rebalance strategy to current situation.

        Parameters
        ----------
        index
            Current date for applying rebalance.

        assets
            Current assets for applying rebalance.

        cash
            Current cash for applying rebalance.

        weights
            weights of assets

        Returns
        -------
        Deal
        """
        return False



"""Rebalance strategies."""
import numpy as np

from mypo.rebalancer.base_rebalancer import BaseRebalancer
from mypo.trigger.no_trigger import NoTrigger


class NoRebalancer(BaseRebalancer):
    """Simple weighted rebalance strategy."""

    def __init__(self) -> None:
        """
        Construct object.

        Parameters
        ----------
        weights
            Weight for applying rebalance.
        """
        super().__init__(trigger=NoTrigger(), weights=np.zeros(1))
